#!/usr/bin/python3

from flask import Flask, url_for, request, Blueprint, session, redirect
from authlib.integrations.flask_client import OAuth
from authlib.oidc.core import CodeIDToken
from authlib.jose import jwt
from urllib.parse import urljoin
import pprint
import html
import os
import requests
import sys
import uuid

SERVER_SESSIONS = dict()


app = Flask(__name__)

app.secret_key = "mysecretkey"

oauth = OAuth(app)

ID_BROKER_KEYCLOAK_FQDN = os.environ.get("ID_BROKER_KEYCLOAK_FQDN")
CLIENT_ID = os.environ.get("CLIENT_ID")
CLIENT_SECRET = os.environ.get("CLIENT_SECRET")
APPLICATION_ROOT = os.environ.get("APPLICATION_ROOT", "")
EXTERNAL_ROOT_URL = os.environ.get("EXTERNAL_ROOT_URL", None)
CLIENT_IP = os.environ.get("CLIENT_IP", None)

bp = Blueprint('uta', __name__)

if not ID_BROKER_KEYCLOAK_FQDN or not CLIENT_ID or not CLIENT_SECRET:
    raise ValueError("Not all env set, check: ID_BROKER_KEYCLOAK_FQDN, CLIENT_ID, CLIENT_SECRET")

oauth.register(
    name="pkce",
    client_id=CLIENT_ID,
    client_secret=CLIENT_SECRET,
    code_challenge_method="S256",
    scope="openid",  # This returns an id token
    server_metadata_url=f"https://{ID_BROKER_KEYCLOAK_FQDN}/auth/realms/ID-Broker/.well-known/openid-configuration"
)

oauth.register(
    name="no_pkce",
    client_id=CLIENT_ID,
    client_secret=CLIENT_SECRET,
    access_token_url='https://{ID_BROKER_KEYCLOAK_FQDN}/auth/realms/ID-Broker/protocol/openid-connect/token',
    access_token_params=None,
    authorize_url='https://{ID_BROKER_KEYCLOAK_FQDN}/auth/realms/ID-Broker/protocol/openid-connect/auth',
    authorize_params=None,
)


def get_full_url(root_url, path):
    if EXTERNAL_ROOT_URL:
        return urljoin(EXTERNAL_ROOT_URL, path)
    else:
        return urljoin(root_url, path)


def get_oauth_session():
    return SERVER_SESSIONS[session["univention_kc"]]


def check_connection():
    try:
        requests.get(f"https://{ID_BROKER_KEYCLOAK_FQDN}")
    except Exception as exc:
        print(f"Could not reach https://login.{ID_BROKER_KEYCLOAK_FQDN}")
        print("Check if it can be resolved and add it to /etc/hosts if necessary")
        print("Check that the CA is available.")
        print("You can get it from \"http://$PRIMARY/ucs-root-ca.crt\"")
        print("Use the env \"REQUESTS_CA_BUNDLE\" and set it to the downloaded CA")
        print(exc)
        sys.exit(1)


@bp.route("/")
def index():
    kc_idp_hint = html.escape(request.args.get("kc_idp_hint", ""), quote=True)
    pkce = html.escape(request.args.get("pkce", ""), quote=True)
    return f"""
Hello<br>
    <form action={url_for(".login")}>
        <label>From which school do you want to login?
            <input type="text" name="kc_idp_hint" value="{kc_idp_hint}">
        </label><br>
        <label>Activate PKCE? [y/n]
            <input type="text" name="pkce" value="{pkce}">
        </label><br>
        <button id=login>"Login"</button>
    </form>
"""


@bp.route("/login")
def login():
    session["kc_idp_hint"] = request.args.get("kc_idp_hint", None)
    session["pkce"] = request.args.get("pkce", "").lower()
    session["univention_kc"] = uuid.uuid4()
    if not session["kc_idp_hint"]:
        raise Exception("No kc_idp_hint request parameter")
    redirect_url = get_full_url(request.url_root, url_for(".authorize"))

    if session["pkce"] == "y":
        print("pkce")
        SERVER_SESSIONS[session["univention_kc"]] = oauth.pkce
    elif session["pkce"] == "n":
        print("no_pkce")
        SERVER_SESSIONS[session["univention_kc"]] = oauth.no_pkce
    else:
        raise Exception("pkce must be 'y' or 'n'")

    return get_oauth_session().authorize_redirect(redirect_url, kc_idp_hint=session["kc_idp_hint"])


@bp.route("/update_tokens")
def update_tokens():
    redirect_url = get_full_url(request.url_root, url_for(".authorize"))
    return get_oauth_session().authorize_redirect(redirect_url, kc_idp_hint=session["kc_idp_hint"])


@bp.route("/authorize")
def authorize():
    # ONLY EVER RETURN THE TOKEN FOR DEBUGGING!!!
    if not session["univention_kc"]:
        return "No oauth state found. Why are you here?"
    if "error" in request.args:
        return f"""INVALID REQUEST<br><pre style="white-space: pre-wrap; word-wrap: break-word;">{html.escape(pprint.pformat(request.args), quote=True)}</pre>"""
    auth_resp = get_oauth_session().authorize_access_token()
    id_claims = jwt.decode(auth_resp['id_token'], get_oauth_session().fetch_jwk_set(), claims_cls=CodeIDToken)
    access_token = jwt.decode(auth_resp["access_token"], key=get_oauth_session().fetch_jwk_set())
    id_claims.validate()

    _id = id_claims['sub']

    return f"""
Success<br>
<a href={url_for(".update_tokens")}>update tokens</a>
<a href={url_for(".logout")}>logout</a>
<h2>Pseudonym:</h2>
<p id=pseudonym>{_id}</p>
<h1>Tokens FOR DEBUGGING ONLY!</h1>
<h2>Token endpoint response:</h2>
<pre style="white-space: pre-wrap; word-wrap: break-word;">{html.escape(pprint.pformat(auth_resp), quote=True)}</pre><br>
<h2>Encoded access Token:</h2>
<pre style="white-space: pre-wrap; word-wrap: break-word; background-color: lightyellow">{html.escape(pprint.pformat(auth_resp["access_token"]), quote=True)}</pre><br>
<h2>Decoded access token</h2>
<pre style="white-space: pre-wrap; word-wrap: break-word;">{html.escape(pprint.pformat(access_token), quote=True)}</pre><br>
<h2>Encoded id token:</h2>
<pre style="white-space: pre-wrap; word-wrap: break-word;">{html.escape(pprint.pformat(auth_resp['id_token']), quote=True)}</pre><br>
<h2>Decoded id token:</h2>
<pre style="white-space: pre-wrap; word-wrap: break-word;">{html.escape(pprint.pformat(id_claims), quote=True)}</pre><br>
<h2>self-disclosure-api curl example:</h2>
 <a href={url_for(".logout")}>logout</a>
"""


@bp.route("/logout")
def logout():
    try:
        SERVER_SESSIONS.pop(session.pop("univention_kc", None), None)
    except KeyError:
        pass
    return redirect(url_for(".index", kc_idp_hint=session["kc_idp_hint"], pkce=session["pkce"]))


if __name__ == "__main__":
    check_connection()
    app.register_blueprint(bp, url_prefix=APPLICATION_ROOT)
    app.run(debug=True, port=5000, host=(CLIENT_IP if CLIENT_IP else "0.0.0.0"))
